from collections.abc import Iterable
from enum import Enum

from phylogenie.tree import Tree
from phylogenie.treesimulator.events import get_mutation_id
from phylogenie.treesimulator.model import get_node_state
from phylogenie.utils import get_heights, get_n_tips, get_times


def _get_states(tree: Tree) -> dict[str, str]:
    return {node.name: get_node_state(node.name) for node in tree}


def _get_mutations(tree: Tree) -> dict[str, int]:
    return {node.name: get_mutation_id(node.name) for node in tree}


class Feature(str, Enum):
    STATE = "state"
    MUTATION = "mutation"
    N_TIPS = "n_tips"
    TIME = "time"
    HEIGHT = "height"


FEATURES_EXTRACTORS = {
    Feature.STATE: _get_states,
    Feature.MUTATION: _get_mutations,
    Feature.N_TIPS: get_n_tips,
    Feature.TIME: get_times,
    Feature.HEIGHT: get_heights,
}


def set_features(tree: Tree, features: Iterable[Feature]) -> None:
    for feature in features:
        feature_maps = FEATURES_EXTRACTORS[feature](tree)
        for node in tree:
            node.set(feature.value, feature_maps[node.name])



from phylogenie.tree import Tree


def get_n_tips(tree: Tree) -> dict[str, int]:
    n_tips: dict[str, int] = {}
    for node in tree.postorder_traversal():
        n_tips[node.name] = (
            1 if node.is_leaf() else sum(n_tips[child.name] for child in node.children)
        )
    return n_tips


def get_times(tree: Tree) -> dict[str, float]:
    times: dict[str, float] = {}
    for node in tree:
        parent_time = 0 if node.parent is None else times[node.parent.name]
        times[node.name] = node.time_to_parent + parent_time
    return times


def get_heights(tree: Tree) -> dict[str, int]:
    heights: dict[str, int] = {}
    for node in tree.postorder_traversal():
        if node.is_leaf():
            heights[node.name] = 0
        else:
            heights[node.name] = 1 + max(heights[child.name] for child in node.children)
    return heights
